@@ -12,7 +12,7 @@ from typing import MutableMapping, List, Union, Coroutine, Type, Sequence, Dict,
 from patchwork.core import Task, AsyncSubscriber
 from patchwork.core.config.base import ComponentConfig
 from patchwork.core.typing import FuncPath
-from patchwork.core.utils import cached_property
+from patchwork.core.utils import cached_property, Flag
 from patchwork.node.core import Module
 from patchwork.node.core.exceptions import TaskControlException, TaskRetry, TaskFatal, TaskDrop, TaskNoHandler
 from patchwork.node.core.router import TaskRouter
@@ -83,10 +83,13 @@ class BaseExecutor(Module):
             FuncPath(check_task_middleware)
         ]
 
+    EVENTS = Module.EVENTS + ('task', )
+
     logger_name = 'executor'
     _loop_task: asyncio.Future
     _routers: List[TaskRouter]
     _unit_monitor_task: asyncio.Future
+    _busy: Flag
 
     def __init__(self, *, parent, **options):
         super().__init__(parent=parent, **options)
@@ -101,6 +104,7 @@ class BaseExecutor(Module):
             self.add_middleware(middleware)
 
         self.add_exception_handler(Exception, default_task_exception_handler)
+        self._busy = Flag()
 
     def __repr__(self):
         res = super().__repr__()
@@ -110,18 +114,33 @@ class BaseExecutor(Module):
     def get_state(self) -> MutableMapping:
         return {
             'worker': self.unit.get_state(),
-            'current_tasks': len(self._current_tasks)
+            'current_tasks': len(self._current_tasks),
+            'busy': self._busy.value
         }
 
+    @property
+    def busy(self) -> Flag:
+        return self._busy
+
     @cached_property
     def unit(self) -> ProcessingUnit:
         return self.settings.unit.instantiate(parent=self)
 
     def add_router(self, router: TaskRouter):
+        """
+        Adds a router to the executor. Routers are responsible for routing tasks to proper handlers.
+        :param router:
+        :return:
+        """
         router.bind(self.worker)
         self._routers.append(router)
 
     def add_middleware(self, middleware: MiddlewareType):
+        """
+        Adds a middleware to be executed for each task before it starts processing
+        :param middleware:
+        :return:
+        """
         self._middlewares.append(middleware)
 
     def add_exception_handler(self, exception: type, handler: ExceptionHandler, idempotent: bool = False):
@@ -241,6 +260,7 @@ class BaseExecutor(Module):
         # future of termination request
         term_future = asyncio.create_task(self._terminate_request.wait())
         subscriber = self.worker.get_subscriber()
+        self._busy.set(False)
 
         while True:
 
@@ -284,19 +304,19 @@ class BaseExecutor(Module):
         task: Task = future.result()
         self.logger.debug(f"Client {receiver} waiter returned a task message: {task.uuid}")
 
+        self._busy.set(True)
+        task_future = self.app_loop.create_future()
+        self._current_tasks.add(task_future)
+        task_future.add_done_callback(self._task_handling_completed)
+
+        await self._dispatch_event('task', task=task)
+
         try:
-            await self.handle(task, receiver)
+            await self.handle(task, receiver, task_future)
         except Exception as e:
-            # log error and rollback task if something went wrong during internal scheduling
+            # log error and do not commit task if something went wrong during internal scheduling
             self.logger.error(f"Unable to handle task '{task.uuid}', exception raised: {e.__class__.__name__}({e})")
-            try:
-                await self.retry(task)
-                # TODO: commit or not?
-            except Exception as e:
-                self.logger.error(f"Unable to retry unhandled task '{task}', "
-                                  f"exception raised: {e.__class__.__name__}({e})")
-                await self.backoff_task(task, reason="Internal retry failed")
-                # TODO: commit or not?
+            task_future.set_exception(e)
 
         return True
 
@@ -350,7 +370,7 @@ class BaseExecutor(Module):
                         extra={'task': task})
         self.logger.debug(f"Task '{task.uuid}' backoffed.{(' Reason: ' + reason) if reason is not None else ''}")
 
-    async def handle(self, task: Task, receiver: AsyncSubscriber) -> Union[asyncio.Future, None]:
+    async def handle(self, task: Task, receiver: AsyncSubscriber, task_future: asyncio.Future):
         """
         Executes given task, handles all exceptions which may rise during task
         processing.
@@ -363,13 +383,11 @@ class BaseExecutor(Module):
             is internally scheduled on processing unit. If you need task execution result await on returned
             future.
 
-        :param task:
-        :param receiver:
-        :return: future of task processing function or None if processing was not scheduled because of
-                 task control exception
+        :param task: task to process
+        :param receiver: a subscriber (receiver) which receives the task
+        :param task_future: a future associated with the task which should resolved when task processing completes
         """
         finalizer = self._handle_task_result
-        task_future = self.app_loop.create_future()
 
         try:
             for middleware in self._middlewares:
@@ -386,10 +404,10 @@ class BaseExecutor(Module):
             partial(self._task_done_callback, finalizer=finalizer(processing_fut=exec_fut), task_future=task_future)
         )
 
-        self._current_tasks.add(task_future)
-        task_future.add_done_callback(lambda f: self._current_tasks.remove(task_future))
-
-        return task_future
+    def _task_handling_completed(self, task_future):
+        self._current_tasks.remove(task_future)
+        if not self._current_tasks:
+            self._busy.set(False)
 
     async def _execute(self, task: Task, receiver: AsyncSubscriber):
         """
